﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SlimDX;

namespace Models.Aqo
{
    public class GetVector3
    {
        public float                    Magnify;
        public Func<Vector3, float>     X;
        public Func<Vector3, float>     Y;
        public Func<Vector3, float>     Z;
        public Func<Vector3, float>     U;
        public Func<Vector3, float>     V;

        public GetVector3(string x, string y, string z, bool exp)
        {
            Init(x, y, z, 1.0f, exp);
        }

        public GetVector3(string x, string y, string z, float mag, bool exp)
        {
            Init(x, y, z, mag, exp);
        }

        public GetVector3(string u, string v, bool exp)
        {
            Init(u, v, exp);
        }

        private void Init(string x, string y, string z, float mag, bool exp)
        {
            Magnify = exp ? mag : 1 / mag;

            switch(x)
            {
            case "+X":  if(exp) X= GetPositiveX; else X= GetPositiveX; break;
            case "-X":  if(exp) X= GetNegativeX; else X= GetNegativeX; break;
            case "+Y":  if(exp) X= GetPositiveY; else Y= GetPositiveX; break;
            case "-Y":  if(exp) X= GetNegativeY; else Y= GetNegativeX; break;
            case "+Z":  if(exp) X= GetPositiveZ; else Z= GetPositiveX; break;
            case "-Z":  if(exp) X= GetNegativeZ; else Z= GetNegativeX; break;
            default:    throw new Exception();
            }

            switch(y)
            {
            case "+X":  if(exp) Y= GetPositiveX; else X= GetPositiveY; break;
            case "-X":  if(exp) Y= GetNegativeX; else X= GetNegativeY; break;
            case "+Y":  if(exp) Y= GetPositiveY; else Y= GetPositiveY; break;
            case "-Y":  if(exp) Y= GetNegativeY; else Y= GetNegativeY; break;
            case "+Z":  if(exp) Y= GetPositiveZ; else Z= GetPositiveY; break;
            case "-Z":  if(exp) Y= GetNegativeZ; else Z= GetNegativeY; break;
            default:    throw new Exception();
            }

            switch(z)
            {
            case "+X":  if(exp) Z= GetPositiveX; else X= GetPositiveZ; break;
            case "-X":  if(exp) Z= GetNegativeX; else X= GetNegativeZ; break;
            case "+Y":  if(exp) Z= GetPositiveY; else Y= GetPositiveZ; break;
            case "-Y":  if(exp) Z= GetNegativeY; else Y= GetNegativeZ; break;
            case "+Z":  if(exp) Z= GetPositiveZ; else Z= GetPositiveZ; break;
            case "-Z":  if(exp) Z= GetNegativeZ; else Z= GetNegativeZ; break;
            default:    throw new Exception();
            }
        }

        private void Init(string u, string v, bool exp)
        {
            switch(u)
            {
            case "U":   if(exp) U= GetU;         else U= GetU;         break;
            case "1-U": if(exp) U= GetInvertU;   else U= GetInvertU;   break;
            case "V":   if(exp) U= GetV;         else V= GetU;         break;
            case "1-V": if(exp) U= GetInvertV;   else V= GetInvertU;   break;
            default:    throw new Exception();
            }

            switch(v)
            {
            case "U":   if(exp) V= GetU;         else U= GetV;         break;
            case "1-U": if(exp) V= GetInvertU;   else U= GetInvertV;   break;
            case "V":   if(exp) V= GetV;         else V= GetV;         break;
            case "1-V": if(exp) V= GetInvertV;   else V= GetInvertV;   break;
            default:    throw new Exception();
            }
        }

        public Vector3 GetPos(Vector3 v)
        {
            return new Vector3(X(v)*Magnify, Y(v)*Magnify, Z(v)*Magnify);
        }

        public Vector3 GetNrm(Vector3 v)
        {
            return new Vector3(X(v), Y(v), Z(v));
        }

        public Vector3 GetUV(Vector3 v)
        {
            return new Vector3(U(v), V(v), 0);
        }

        public static float GetPositiveX(Vector3 v)    { return  v.X; }
        public static float GetNegativeX(Vector3 v)    { return -v.X; }
        public static float GetPositiveY(Vector3 v)    { return  v.Y; }
        public static float GetNegativeY(Vector3 v)    { return -v.Y; }
        public static float GetPositiveZ(Vector3 v)    { return  v.Z; }
        public static float GetNegativeZ(Vector3 v)    { return -v.Z; }
        public static float GetU(Vector3 v)            { return  v.X; }
        public static float GetInvertU(Vector3 v)      { return 1-v.X; }
        public static float GetV(Vector3 v)            { return  v.Y; }
        public static float GetInvertV(Vector3 v)      { return 1-v.Y; }
    }

    public class GeometryUtils
    {
        public static void VerifyTriangleList(List<Triangle> a, List<Triangle> b)
        {
            foreach(var i in a)
            {
                int find    = -1;

                for(int j= 0; j < b.Count; ++j)
                {
                    var t   = b[j];

                    if(i[0] == t[0] && i[1] == t[1] && i[2] == t[2])
                    {
                        find    = j;
                        break;
                    }

                    if(i[0] == t[1] && i[1] == t[2] && i[2] == t[0])
                    {
                        find    = j;
                        break;
                    }

                    if(i[0] == t[2] && i[1] == t[0] && i[2] == t[1])
                    {
                        find    = j;
                        break;
                    }
                }

                AFPFile.Assert(find != -1);

                b.RemoveAt(find);
            }

            AFPFile.Assert(b.Count == 0);
        }

        public static Triangle GetConnect(List<Triangle> triangles, ushort a, ushort b)
        {
            for(int i= 0; i < triangles.Count; ++i)
            {
                var t   = triangles[i];

                if(t[0] == a && t[1] == b)
                {
                    triangles.RemoveAt(i);
                    return t;
                  //return new ushort[] { a, b, t[2] };
                }

                if(t[1] == a && t[2] == b)
                {
                    triangles.RemoveAt(i);
                    return new Triangle(a, b, t[0]);
                }

                if(t[2] == a && t[0] == b)
                {
                    triangles.RemoveAt(i);
                    return new Triangle(a, b, t[1]);
                }
            }

            return null;
        }

        public static IEnumerable<Triangle> CreateTriangleList(ushort[] indices)
        {
            for(int i= 0; i < indices.Length; i+=3)
                yield return new Triangle(indices[i+0], indices[i+1], indices[i+2]);
        }

        public static IEnumerable<Triangle> CreateTriangleStrip(ushort[] indices)
        {
            var a   = indices[0];
            var b   = indices[1];

            for(int i= 2; i < indices.Length; ++i)
            {
                var c   = indices[i];

                if(a != b && b != c && c != a)
                    yield return (i % 2) == 0 ? new Triangle(a, b, c) : new Triangle(a, c, b);

                a   = b;
                b   = c;
            }
        }

        public static ushort[] Stripify(ushort[] indices)
        {
            var triangles   = CreateTriangleList(indices).ToList();
            var newindices  = new List<ushort>();

            newindices.Add(triangles[0][0]);
            newindices.Add(triangles[0][1]);
            newindices.Add(triangles[0][2]);

            triangles.RemoveAt(0);

            while(triangles.Count > 0)
            {
                var n   = newindices.Count - 2;

                if((n & 1) == 0)
                {
                    var t   = GetConnect(triangles, newindices[n], newindices[n+1]);

                    if(null == t)
                    {
                        t   = triangles[0];

                        newindices.Add(newindices[n+1]);
                        newindices.Add(t[0]);
                        newindices.Add(t[0]);
                        newindices.Add(t[1]);
                        newindices.Add(t[2]);
                        triangles.RemoveAt(0);
                    } else
                    {
                        newindices.Add(t[2]);

                        AFPFile.Assert(newindices[newindices.Count - 3] == t[0]
                                    && newindices[newindices.Count - 2] == t[1]
                                    && newindices[newindices.Count - 1] == t[2]);
                    }
                } else
                {   // 反転
                    var t   = GetConnect(triangles, newindices[n+1], newindices[n]);

                    if(null == t)
                    {
                        t   = triangles[0];

                        newindices.Add(newindices[n+1]);
                        newindices.Add(t[0]);
                        newindices.Add(t[0]);
                        newindices.Add(t[2]);
                        newindices.Add(t[1]);
                        triangles.RemoveAt(0);
                    } else
                    {
                        newindices.Add(t[2]);

                        AFPFile.Assert(newindices[newindices.Count - 3] == t[1]
                                    && newindices[newindices.Count - 2] == t[0]
                                    && newindices[newindices.Count - 1] == t[2]);
                    }
               }
            }

            return newindices.ToArray();
        }

        public static void CalcTangentAndBinormal(
            Vector3 p0, Vector3 uv0,
            Vector3 p1, Vector3 uv1,
            Vector3 p2, Vector3 uv2,
            out Vector3 t, out Vector3 b)
        {
            var cp0 = new [] { new Vector3(p0.X, uv0.X, uv0.Y), new Vector3(p0.Y, uv0.X, uv0.Y), new Vector3(p0.Z, uv0.X, uv0.Y) };
            var cp1 = new [] { new Vector3(p1.X, uv1.X, uv1.Y), new Vector3(p1.Y, uv1.X, uv1.Y), new Vector3(p1.Z, uv1.X, uv1.Y) };
            var cp2 = new [] { new Vector3(p2.X, uv2.X, uv2.Y), new Vector3(p2.Y, uv2.X, uv2.Y), new Vector3(p2.Z, uv2.X, uv2.Y) };
            var u   = new float[3];
            var v   = new float[3];

            for(int i= 0; i < 3; ++i)
            {
                var v1  = cp1[i] - cp0[i];
                var v2  = cp2[i] - cp1[i];
                var c   = Vector3.Cross(v1, v2);

                //AFPFile.Assert(v1.LengthSquared() != 0);
                //AFPFile.Assert(v2.LengthSquared() != 0);

                if(c.X == 0)
                {
                    //AFPFile.Assert(false);

                    t   = Vector3.Zero;
                    b   = Vector3.Zero;
                    return;
                }

                u[i]= - c.Y / c.X;
                v[i]= - c.Z / c.X;
            }

            t       = Vector3.Normalize(new Vector3(u[0], u[1], u[2]));
            b       = Vector3.Normalize(new Vector3(v[0], v[1], v[2]));
        }
    }

    public class Triangle
    {
        public ushort[]                 Indices                 { get; set; }
        public ushort                   this[int index]         { get { return Indices[index]; } }

        public Triangle(params ushort[] indices)
        {
            Indices = indices;
        }

        public override string ToString()
        {
            return string.Format("{0} {1} {2}", Indices[0], Indices[1], Indices[2]);
        }
    }
}
